"""Constants for AnonCreds."""

DEFAULT_CRED_DEF_TAG = "default"
DEFAULT_MAX_CRED_NUM = 1000
DEFAULT_SIGNATURE_TYPE = "CL"

CATEGORY_SCHEMA = "schema"

CATEGORY_CRED_DEF = "credential_def"
CATEGORY_CRED_DEF_PRIVATE = "credential_def_private"
CATEGORY_CRED_DEF_KEY_PROOF = "credential_def_key_proof"

CATEGORY_REV_LIST = "revocation_list"
CATEGORY_REV_REG_DEF = "revocation_reg_def"
CATEGORY_REV_REG_DEF_PRIVATE = "revocation_reg_def_private"

STATE_FINISHED = "finished"
STATE_REVOCATION_PENDING = "pending"
STATE_REVOCATION_POSTED = "posted"

REV_REG_DEF_STATE_ACTIVE = "active"
